from datetime import timedelta
import pandas as pd
from sinamet.tools.timext import to_date
from sinamet.tools.listmtq import (list_flows_to_dataframe,
                                   dataframe_flows_to_timeval,
                                   dataframe_flows_to_timeval_freq,
                                   dataframe_flows_to_val)


class Profile():
    """
    Série de valeurs datées.

    Parameters:
        times: liste des repères temporels (dates).
        values: liste des valeurs associées à chaque repère temporel.
    """
    def __init__(self, times=[], values=[], properties=None):
        if type(times) is not list:
            raise TypeError("Excepted list for 'times' attribute")
        if len(times) != len(values):
            raise ValueError("'times' and 'values' must have same length")
        if properties is not None:
            df = pd.DataFrame({"property": properties})
            df["date_start"] = df["property"].apply(lambda a: a.date_start)
            df["date_end"] = df["property"].apply(lambda a: a.date_end + timedelta(days=1))
            df["value"] = df["property"].apply(lambda a: a.get_value())
            df = df.sort_values(by=['date_start'], ascending=True)
            self.times = df["date_start"].to_list() + [df["date_end"].to_list()[-1]]
            self.values = df["value"].to_list() + [0.0]
        else:
            self.times = [to_date(t) for t in times]
            self.values = values

    def __str__(self):
        return f"<Profile len={len(self.times)}>"

    def __repr__(self):
        return f"<<PROFILE len={len(self.times)}----{self.times}----{self.values}>>"

    def isempty(self):
        return len(self.times) == 0

    def iszero(self):
        if self.isempty():
            return True
        if max(self.values) == min(self.values) == 0:
            return True
        return False

    def trunc(self, date_start="01/01/1900", date_end="31/12/2100"):
        date_start = to_date(date_start)
        date_end = to_date(date_end)
        tb = []
        vb = []
        for t, v in zip(self.times, self.values):
            if t >= date_start and t < date_end:
                tb.append(t)
                vb.append(v)
            if t >= date_end:
                tb.append(t)
                vb.append(v)
                break

        return Profile(times=tb, values=vb)

    @staticmethod
    def build(lst_qt, unit, freq=None, per_day=True):
        """
        Construire un Profile avec une liste de flux et une unité spécifiée.

        Parameters:
            lst_qt: List de quantifiables (flux)
            unit: Unité souhaitée
            freq:
            per_day: valeur quotidienne
        """
        if len(lst_qt) == 0:
            return Profile([], [])
            # return Profile([to_date(d) for d in defdates], [0]*len(defdates))
        (t, v) = Profile.static_get_profile(lst_qt, unit, freq=freq, per_day=per_day)
        return Profile(t, v)

    @staticmethod
    def evalstr(my_str):
        lst = my_str.split("----")
        t = eval(lst[1].strip())
        v = eval(lst[2].replace(">>", "").strip())
        return Profile(t, v)

    @staticmethod
    def homogeneize(*lst_profiles):
        temp_prof = lst_profiles[0]
        if type(temp_prof) is list:
            temp_prof = temp_prof[0]
            lst_profiles = lst_profiles[0]
        for prof in lst_profiles[1:]:
            t, vtemp, v2 = Profile.static_operation(temp_prof.times, temp_prof.values,
                                                    prof.times, prof.values, '=')
            temp_prof = Profile(times=t, values=vtemp)
        lst_return = []
        for prof in lst_profiles:
            t, vtemp, v2 = Profile.static_operation(temp_prof.times, temp_prof.values,
                                                    prof.times, prof.values, '=')
            lst_return.append(Profile(times=t, values=v2))
        return lst_return

    @staticmethod
    def static_get_profile(flows, unit, freq=None, per_day=True, step_profile=True):
        """
        Return (time, val) profile for a liste of quantifiable
        """

        if type(flows) is list:
            _df = list_flows_to_dataframe(flows, unit, per_day=per_day)
        elif flows.__class__.__name__ == "DataFrame":
            _df = flows
        else:
            raise TypeError("Unknown type : " + str(type(flows)))

        if freq is None or freq == "":
            (t, v) = dataframe_flows_to_timeval(_df, unit, per_day=per_day)
        else:
            (t, v) = dataframe_flows_to_timeval_freq(_df, unit, freq, per_day=per_day)

        if type(unit) is list:
            if step_profile:
                for key, val in v.items():
                    v[key] = val[:-1] + [val[-2]]
        else:
            if step_profile:
                if type(v) is dict:
                    v = v[unit]
                try:
                    v = v[:-1] + [v[-2]]
                except IndexError:
                    print("TIME ERROR = %s" % t)
                    print("VAL ERROR = %s" % v)
                    raise

        return t, v

    @staticmethod
    def get_value(flows, unit, start=None, end=None, year=None, month=None):
        """
        Renvoie la valeur d'une liste de flux sur la période indiquée.
        La période est renseignée soit par une date de début et/ou de fin,
        soit par une année, soit par une année et un mois (de 1 à 12)

        Parameters:
            flows: Liste des flux
            unit: Unité pour exprimer les flux
            start: Date de début
            end: Date de fin
            year: Année
            month: Mois
        """
        # CONVERSION LIST FLOW INTO DATAFRAME
        if type(flows) is list:
            if len(flows) == 0:
                return 0.0
            _df = list_flows_to_dataframe(flows, unit)
        elif flows.__class__.__name__ == "DataFrame":
            if len(flows.index) == 0:
                return 0.0
            _df = flows
        else:
            raise TypeError("Unknown type : " + str(type(flows)))

        # CONVERTIR
        _returnval = dataframe_flows_to_val(
            _df, unit, start=start, end=end, year=year, month=month)
        # print("Return val = " + str(_returnval))
        # print("UNIT = " + str(unit))
        if type(unit) is list:
            # print("Unit is list !!! Return = " + str(_returnval))
            return _returnval
        else:
            # print("Unit is NOT list !!! Return = " + str(_returnval[unit]))
            return _returnval[unit]

    def __mul__(self, other):
        if type(other) is int or type(other) is float:
            v = [val*other for val in self.values]
            return Profile(self.times, v)
        else:
            t, v = self.compute(other, "*")
            return Profile(t, v)

    def __truediv__(self, other):
        if type(other) is int or type(other) is float:
            v = [val/other for val in self.values]
            return Profile(self.times, v)
        elif type(other) is Profile:
            t, v = self.compute(other, "/")
            return Profile(t, v)
        else:
            print("ERROR CONTEXT")
            raise ValueError("Cannot __truediv__ with '%s' [type=%s]" % (other, type(other)))

    def __add__(self, other):
        if type(other) is int or type(other) is float:
            v = [val + other for val in self.values]
            return Profile(self.times, v)
        else:
            t, v = self.compute(other, "+")
            return Profile(t, v)

    def __sub__(self, other):
        t, v = self.compute(other, "-")
        return Profile(t, v)

    def max(self, default=0):
        if self.isempty():
            return default
        return max(self.values)

    def integrate(self):
        sumval = 0
        listval = []
        for val in self.values:
            sumval += val
            listval.append(sumval)
        return Profile(self.times, listval)

    def compute(self, profile, operator):
        return Profile.static_operation(self.times, self.values,
                                        profile.times, profile.values, operator)

    @staticmethod
    def static_operation(time1, val1, time2, val2, operator):
        rtime = []
        rval = []
        if operator == "=":
            rval2 = []

        for t in time1:
            rtime.append(t)
            rval.append(0)
            if operator == "=":
                rval2.append(0)

        for t in time2:
            if t not in rtime:
                rtime.append(t)
                rval.append(0)
                if operator == "=":
                    rval2.append(0)

        rtime.sort()

        for i in range(0, len(time1) - 1):
            index_start = rtime.index(time1[i])
            index_end = rtime.index(time1[i + 1])

            for j in range(index_start, index_end):
                rval[j] += val1[i]

        for i in range(0, len(time2) - 1):
            index_start = rtime.index(time2[i])
            index_end = rtime.index(time2[i + 1])

            for j in range(index_start, index_end):
                if operator == "/":
                    try:
                        rval[j] = rval[j] / val2[i]
                    except ZeroDivisionError:
                        rval[j] = 0.0001
                elif operator == "*":
                    rval[j] = rval[j] * val2[i]
                elif operator == "+":
                    rval[j] = rval[j] + val2[i]
                elif operator == "-":
                    rval[j] = rval[j] - val2[i]
                elif operator == "=":
                    rval2[j] = val2[i]

        if operator == "=":
            return rtime, rval, rval2

        return rtime, rval

    @staticmethod
    def build_test():
        return Profile(times=["31/12/2019", "31/12/2020", "31/12/2021", "31/12/2022"],
                       values=[10, 5, 8, 3])



from sqlalchemy.schema import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship

from .quantifiable import MTQuantifiable
from .actor import Actor
from .territory import Territory


class Pathflow(MTQuantifiable):
    """Table flux de chemin."""
    __tablename__ = 'pathflow'

    id: Mapped[int] = mapped_column(ForeignKey('mtquantifiable.id', ondelete='CASCADE'),
                                    primary_key=True)

    emitter_actor_id: Mapped[int | None] = mapped_column(ForeignKey('actor.id'))
    emitter_actor: Mapped[Actor | None] = relationship(foreign_keys=emitter_actor_id)

    emitter_territory_id: Mapped[int | None] = mapped_column(ForeignKey('territory.id'))
    emitter_territory: Mapped[Territory | None] = relationship(foreign_keys=emitter_territory_id)

    receiver_actor_id: Mapped[int | None] = mapped_column(ForeignKey('actor.id'))
    receiver_actor: Mapped[Actor | None] = relationship(foreign_keys=receiver_actor_id)

    receiver_territory_id: Mapped[int | None] = mapped_column(ForeignKey('territory.id'))
    receiver_territory: Mapped[Territory | None] = relationship(foreign_keys=receiver_territory_id)

    __mapper_args__ = {
        'polymorphic_identity': 'pathflow',
    }



import pandas

from datetime import timedelta, date, datetime
import numpy as np
from .timext import full_date_range, common_nb_days, get_start_end_dates

#import swifter


def list_flows_to_dataframe(list_flows, unity, per_day = True):
    df = list_periodic_to_dataframe(list_flows)

    if type(unity) is not list:
        unity = [unity]
    for unit in unity:
        #df[unit] = parallelize(df['object'], get_quantity(unit)
        df[unit] = df['object'].apply(lambda v: v.get_quantity(unit))
        if per_day:
            df[unit + "_per_day"] = df[unit] / df['nb_days']
    return df


def list_periodic_to_dataframe(list_periodic):
    # todo: remove object to use only mtobject
    dic = {'object':list_periodic, 'mtobject':list_periodic}
    df = pandas.DataFrame.from_dict(dic)

    # LIST FLOWS = DATE START + END INCLUES
    # DATAFRAME = START INCLUE, END EXCLUE

    df['date_start']= df['mtobject'].apply(lambda v: date_point_flow_manage(v, "start"))
    df['date_end'] = df['mtobject'].apply(lambda v: date_point_flow_manage(v, "end"))
    """df['date_start'] = df['mtobject'].apply(lambda v: v.date_start)
    df['date_end'] = df['mtobject'].apply(lambda v: v.date_end + timedelta(days=1))"""
    df['nb_days_timedelta'] = df['date_end'] - df['date_start']
    df['nb_days'] = df['nb_days_timedelta'].apply(lambda v: v.days)

    return df

def date_point_flow_manage(mtqobject, dir):
    if mtqobject.date_point is not None:
        dst = mtqobject.date_point
        den = mtqobject.date_point + timedelta(days=1)
    else:
        dst = mtqobject.date_start
        den = mtqobject.date_end + timedelta(days=1)
    if dir == "start":
        return dst
    elif dir == "end":
        return den

def list_datepoint_to_dataframe(list_datepoint):
    # todo: remove object to use only mtobject
    dic = {'mtobject':list_datepoint}
    df = pandas.DataFrame.from_dict(dic)
    df['date_point'] = df['mtobject'].apply(lambda v: v.date_point)
    return df

def dataframe_flows_to_val(df, unity,
                           start = None, end = None,
                           year = None, month = None):
    """start & end dates or str
    End & Start included
    """
    if df.empty:
        return None

    # CONVERSION START & END FROM STRING TO DATE OBJECT
    (start, end) = get_start_end_dates(start = start, end = end,
                                       year = year, month = month)
    if (start, end) == get_start_end_dates():
        raise ValueError("You must specify a period : start/end or year (+month)")
    end = end + timedelta(days=1)
    
    if type(unity) is not list:
        unity = [unity]

    values = {}        
    sub_df = df[(df['date_end'] >= start) &
                (df['date_start'] < end)].copy()

    if not sub_df.empty:
        sub_df['temp_common_days'] = sub_df.apply(lambda row: common_nb_days(\
            row['date_start'], row['date_end'], \
            start, end), axis=1)
        
        for unit in unity:
            sub_df[unit + '_temp_result'] = sub_df['temp_common_days'] *\
                                        sub_df[unit + '_per_day']
            values[unit] = sub_df[unit + '_temp_result'].sum()
        """print(sub_df)
        for i in sub_df.iterrows():
            print(i)"""
    else:
        for unit in unity:
            values[unit] = 0
        
        #print(str(unit) + " : " + str(values[unit][-1]))
        """sub_df.to_excel("../data_out/Conso_bat_out" + str(_listdates[i])\
                    + ".xls")"""
    return values

def dataframe_flows_to_timeval(df, unity, per_day = True):
    start_date_unique = df['date_start'].unique().tolist()
    end_date_unique = df['date_end'].unique().tolist()
    date_lst = sorted(np.unique(start_date_unique + end_date_unique))

    unity_is_list = False
    if type(unity) is not list:
        unity_is_list = True
        unity = [unity]

    #if type(unity) is list:
    values = {}
    for unit in unity:
        values[unit] = []
    for i in range(len(date_lst)-1):
        sub_df = df[(df['date_start']<=date_lst[i]) & (df['date_end']>=date_lst[i+1])]
        for unit in unity:
            if per_day:
                values[unit].append(sub_df[unit + '_per_day'].sum())
            else:
                """print("Longueur résult = %s" % len(sub_df.index))
                print("return = %s"%sub_df[unit].iloc[0])"""
                values[unit].append(sub_df[unit].sum())
    for unit in unity:
        values[unit] += [0]
    """else:
        values = []
        for i in range(len(date_lst)-1): 
            sub_df = df[(df['date_start']<=date_lst[i]) & (df['date_end']>=date_lst[i+1])]
            if per_day:
                values.append(sub_df[unity + '_per_day'].sum())
            else:
                print("Longueur résult = %s" % len(sub_df.index))
                print("return = %s" % sub_df[unity].iloc[0])
                values.append(sub_df[unity].iloc[0])
        values += [0]"""

    if unity_is_list:
        return (date_lst, values)
    else:
        return (date_lst, values[unity[0]])

def dataframe_flows_to_timeval_freq(df, unity, freq, per_day = True):
    if freq is None:
        return dataframe_flows_to_timeval(df, unity)
    """Frequency = pandas.date_range parameter"""
    # Date range =
    # Voir doc : https://pandas.pydata.org/pandas-docs/stable/...
    # ...generated/pandas.date_range.html
    if df.empty:
        return ([],[])

    _start_date = df['date_start'].min()
    _end_date = df['date_end'].max()

    _listdates = full_date_range(_start_date, _end_date, freq)
    
    _listdates = [dt.date() for dt in _listdates]
    #Debug.shell("Date liste = " + str(_listdates))

    unity_is_list = False
    if type(unity) is not list:
        unity_is_list = True
        unity = [unity]

    values = {}
    for unit in unity:
        values[unit] = []

    #print(_listdates)
        
    for i in range(len(_listdates)-1):
        sub_df = df[(df['date_end'] >= _listdates[i]) &
                    (df['date_start'] < _listdates[i+1])].copy()
        #print(_listdates[i], end="")
        if not sub_df.empty:
            sub_df['temp_common_days'] = sub_df.apply(lambda row: common_nb_days(\
                row['date_start'], row['date_end'], \
                _listdates[i], _listdates[i+1]), axis=1)
            
            for unit in unity:
                if per_day:
                    sub_df[unit + '_temp_result'] = sub_df['temp_common_days'] *\
                                                sub_df[unit + '_per_day']
                    #print(sub_df)
                    values[unit].append(sub_df[unit + '_temp_result'].sum())
                else:
                    raise AttributeError("'per_day = False' is not implemented yet")
        else:
            values[unit].append(0)
            #print(str(unit) + " : " + str(values[unit][-1]))
            """sub_df.to_excel("../data_out/Conso_bat_out" + str(_listdates[i])\
                        + ".xls")"""
            
    for unit in unity:
        values[unit] += [0]
   
    if unity_is_list:
        return (_listdates, values)
    else:
        return (_listdates, values[unity[0]])

def extend_time(x1, y1, x2, y2):

    timeline = sorted(np.unique(x1+x2))

    result1 = []
    result2 = []

    #print(x1, y1, x2, y2)

    for t in timeline:
        if t in x1:
            result1.append(y1[x1.index(t)])
        else:
            result1.append(0)
        if t in x2:
            result2.append(y2[x2.index(t)])
        else:
            result2.append(0)
    return (timeline, result1, result2)

    
def cross_xy(x1, y1, x2, y2):  # UNUSED
    """Sorted and compatible L1 & L2"""
    _resultx = []
    _resulty = []
    _timeline = []
    for i in range(len(x1)-1):
        if x1[i] in x2:
            _timeline.append(x1[i])
            _resultx.append(y1[i])
            _resulty.append(y2[x2.index(x1[i])])

    return (_timeline, _resultx, _resulty)





from .mapper import Mapper
from .mapper import MapperError

from .config import config

from .sidb import Sidb

from .mtobjects import MTObject
from .mtobjects import Territory
from .mtobjects import Actor
from .mtobjects import Product
from .mtobjects import Pathflow
from .mtobjects import Gateflow
from .mtobjects import Stock
from .mtobjects import Property

__version__ = '0.1.1'



import hashlib
import base64
import unicodedata
import re


def hexhash(string, length = 30):
    return hashlib.sha224( \
        str(string).encode('utf-8')).hexdigest()[0:length]


def alphanumhash(*content, length=32):
    """Return 0-9a-z hash string"""
    if length > 86:
        raise ValueError("Too high length=%s (max=86)" % length)
    mycontentstr = "/".join([str(c) for c in content])
    mycontent_hashed = str(mycontentstr).encode('utf-8')
    hash_filename = hashlib.sha512(mycontent_hashed).digest()
    myhash = str(base64.b64encode(hash_filename)).lower().replace("/","0").replace("+","1")[2:2 + length]
    return myhash


def minimize(string: str) -> str:
    """Renvoie une chaine sans accents, ni caractères spéciaux, ni espace en miniscule"""
    special_characters: dict[str, str] = {
            'æ': 'ae',
            'œ': 'oe',
            'ß': 'ss',
    }
    nfkd_form = unicodedata.normalize('NFKD', string.lower())
    unicode_lst = [c for c in nfkd_form if not unicodedata.combining(c)]
    unicode_str = u''.join([special_characters[c] if c in special_characters else c
                            for c in unicode_lst])
    return re.sub('[^a-z0-9]', '', unicode_str)



from pathlib import Path

def templatepath(file):
    return Path(file).parent / "template"

def datapath(file):
    return Path(file).parent / "data"


def buildpath(file):
    return Path(file).parent / "build"


def imgpath(file):
    return Path(file).parent / "img"


def cachepath(file):
    return Path(file).parent / "cache"


def buildhtmlpath(file):
    return buildpath(file) / ("%s.html" % Path(file).stem)


def cachehtmlpath(file):
    return cachepath(file) / ("%s.html" % Path(file).stem)


def makepath(file, folder=None, ext=None):
    mypath = Path(file).parent
    if folder is not None:
        mypath = mypath / folder
    if ext is not None:
        if not ext.startswith("."):
            ext = "." + ext
        mypath = mypath / ("%s%s" % (Path(file).stem, ext))
    return mypath


from sqlalchemy.schema import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column

from .mtobject import MTObject


class Product(MTObject):
    """Table produit."""
    __tablename__ = 'product'

    id: Mapped[int] = mapped_column(ForeignKey('mtobject.id', ondelete='CASCADE'),
                                    primary_key=True)

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }

    def __str__(self) -> str:
        code = self.get_code() or 'No code'
        name = self.get_name() or 'No name'
        nomenclature = self.get_property('Nomenclature') or 'No nomenclature'
        return f'<Product-{self.id}: {nomenclature}({code})-{name}>'



from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship

from .mtobject import MTObject
from .territory import Territory


class Actor(MTObject):
    """Table acteur."""
    __tablename__ = 'actor'

    id: Mapped[int] = mapped_column(ForeignKey('mtobject.id', ondelete='CASCADE'),
                                    primary_key=True)

    territory_id: Mapped[int | None] = mapped_column(ForeignKey('territory.id'))
    territory: Mapped[Territory | None] = relationship(foreign_keys=territory_id)

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }

    def __str__(self) -> str:
        return f'<Actor-{self.id}: {self.get_name()}>'



from __future__ import annotations
from datetime import date, datetime
import enum
from typing import Any, TYPE_CHECKING

from geoalchemy2 import Geometry, WKBElement
from geoalchemy2.shape import from_shape, to_shape
import shapely
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import JSON
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship
from sqlalchemy.orm import reconstructor

from sinamet.tools.timext import to_date

from .dbobject import DBObjectClassBase

if TYPE_CHECKING:
    from .mtobject import MTObject


class Property(DBObjectClassBase):
    """Objet Property

     Attributes:
            item: L'objet lié à la propriété.
            source_ref: La source de référence de la propriété.
            date_point: La date ponctuelle de la propriété (si existante).
            date_start: La date de début de la propriété (si existante).
            date_end: La date de fin de la propriété (si existante).
            value_mtobject: La valeur de la propriété si c'est une relation avec un autre MTObject.
            value_literal: La valeur si c'est un nombre ou du texte.
            value_geo: La valeur si c'est une forme géographique.
            context: Dictionnaire complémentaire pour contextualiser la propriété (Optionnel)
            timestamp: Date d'insertion de la propriété dans la base de données.
    """
    class ValueType(enum.Enum):
        MTOBJECT = 1
        STRING = 2
        FLOATING = 3
        INTEGER = 4
        GEOMETRY = 5

    __tablename__ = 'property'

    id: Mapped[int] = mapped_column(primary_key=True)
    type: Mapped[ValueType] = mapped_column(Enum(ValueType))
    name_a: Mapped[str] = mapped_column(index=True)
    name_b: Mapped[str | None] = mapped_column(index=True)
    item_id: Mapped[int] = mapped_column(ForeignKey('mtobject.id',
                                                    ondelete="CASCADE"),
                                         index=True)
    item: Mapped[MTObject] = relationship(back_populates="properties",
                                          foreign_keys=item_id)
    value_mtobject_id: Mapped[int | None] = mapped_column(
            ForeignKey('mtobject.id', ondelete="CASCADE"),
            index=True)
    value_mtobject: Mapped[MTObject | None] = relationship(
            back_populates="property_values",
            foreign_keys=value_mtobject_id)
    value_literal: Mapped[str | None] = mapped_column(index=True)
    value_geo: Mapped[WKBElement | None] = mapped_column(Geometry)
    source_ref: Mapped[str | None] = mapped_column(index=True)
    date_start: Mapped[date | None]
    date_end: Mapped[date | None]
    date_point: Mapped[date | None]
    context: Mapped[dict[str, Any] | None] = mapped_column(JSON)
    timestamp: Mapped[datetime]

    # datalinker_id = Column(Integer, ForeignKey('datalinker.id'))
    # datalinker = relationship("DataLinker", back_populates="attributes")
    # overwrite_id = Column(Integer, ForeignKey('property.id'))
    # overwrite = relationship("Property", remote_side=[id])
    # overwrittenby = relationship("Property", back_populates="overwrite")

    def __init__(self,
                 name: str,
                 value: str | int | float | MTObject | shapely.Geometry,
                 source_ref: str | None = None,
                 date_point: str | date | None = None,
                 date_start: str | date | None = None,
                 date_end: str | date | None = None,
                 context: dict[str, Any] | None = None) -> None:
        """Constructeur de la propriété.

        Parameters:
            name: Nome complet de la propriété.
            value: Valeur de la propriété.
            source_ref: La source de référence de la propriété.
            date_point: La date ponctuelle de la propriété.
            date_start: La date de début de la propriété.
            date_end: La date de fin de la propriété.
            context: Dictionnaire complémentaire pour contextualiser la propriété (Optionnel)

        Raises:
            ValueError: Quand la valeur est vide.
            TypeError: Quand le type de la valeur est incorrect.
        """
        from .mtobject import MTObject

        if value is None or value == '':
            raise ValueError(f'Cannot set a property with an empty value ({value=}).')

        name_a, *name_b = name.split('@', 1)
        self.name_a = name_a
        if name_b:
            self.name_b = name_b[0]

        match value:
            case MTObject():
                self.value_mtobject = value
                self.type = self.ValueType.MTOBJECT
            case str():
                self.value_literal = value
                self.type = self.ValueType.STRING
            case int():
                self.value_literal = str(value)
                self.type = self.ValueType.INTEGER
            case float():
                self.value_literal = str(value)
                self.type = self.ValueType.FLOATING
            case shapely.Geometry():
                self.value_geo = from_shape(value)
                self.type = self.ValueType.GEOMETRY
            case _:
                raise TypeError(f'Invalid Property type ({type(value)})')

        self.source_ref = source_ref
        self.date_start = to_date(date_start)
        self.date_end = to_date(date_end)
        self.date_point = to_date(date_point)
        self.timestamp = datetime.now()
        self.context = context

    @reconstructor
    def init_on_load(self) -> None:
        """Charge la propriété dans le cache de l'objet auquel elle appartient."""
        fullname = self.name_a
        if self.name_b is not None:
            fullname += "@" + self.name_b
        if fullname not in self.item.cached_properties:
            self.item.cached_properties[fullname] = []
        self.item.cached_properties[fullname].append(self)

    def get_value(self) -> (str | int | float | MTObject | shapely.Geometry):
        """Renvoie la valeur de la propriété.
        Peut être un litéral (texte, nombre), un MTObject ou une forme géographique
        """
        if self.type.name == 'MTOBJECT' and self.value_mtobject:
            return self.value_mtobject
        if self.type.name == 'STRING' and self.value_literal:
            return self.value_literal
        if self.type.name == 'FLOATING' and self.value_literal:
            return float(self.value_literal)
        if self.type.name == 'INTEGER' and self.value_literal:
            return int(self.value_literal)
        if self.type.name == 'GEOMETRY' and self.value_geo:
            return to_shape(self.value_geo)
        raise TypeError(f'Invalid Property type ({self.type})')

    def get_name(self) -> str:
        """Renvoie le nom complet de la propriété."""
        return self.name_a + ('' if self.name_b is None else f'@{self.name_b}')

    def get_type(self) -> str:
        """Renvoie le type de la valeur de la propriété."""
        return self.type.name

    def get_value_str(self) -> str:
        """Renvoie la valeur de la propriété sous forme de string."""
        if self.type.name == 'MTOBJECT':
            return str(self.value_mtobject)
        if self.type.name in ('STRING', 'INTEGER', 'FLOATING'):
            return str(self.value_literal)
        return 'Geometry'

    def get_datetime_str(self) -> str:
        """Renvoie la temporalité de la propriété sous forme de string."""
        _return_str = ""
        if self.date_point is not None:
            _return_str += " [" + str(self.date_point) + "]"
        if self.date_start is not None or self.date_end:
            _return_str += " [" + str(self.date_start) + \
                          "->" + str(self.date_end) + "]"
        return _return_str

    def get_source_ref_str(self) -> str | None:
        """Renvoie la source de référence de la propriété."""
        return self.source_ref

    def __str__(self) -> str:
        return (f'{self.get_name()} ({self.get_type()}) - {self.get_value_str()}'
                f' - {self.get_datetime_str()} - {self.get_source_ref_str()}')

    def __repr__(self) -> str:
        return f"<Property:{self.name_a}>"

    def has_date(self) -> bool:
        """Détermine si la propriété est liée à une temporalité."""
        return (self.date_point is not None
                or self.date_start is not None
                or self.date_end is not None)



from __future__ import annotations

import typing
from typing import Any

import warnings

from sinamet.tools.timext import get_start_end_dates, to_date

if typing.TYPE_CHECKING:
    from datetime import date, datetime
    from shapely import Geometry
    from sinamet.mtobjects.mtobject import MTObject


class Mapper():
    def __init__(self,
                 type: str,
                 source_ref: str | None = None
                 ) -> None:
        """Constructeur du Mapper.

        Parameters:
            type: Le type d'objet lié au Mapper (Territory, Actor, ...)
            source_ref: La source de référence de l'objet.
        """
        self.type: str = type.lower()

        self.keys: list[str] = []
        self.values: list[str | MTObject | Geometry] = []
        self.date_start: list[date | None] = []
        self.date_end: list[date | None] = []
        self.date_point: list[date | None] = []
        self.source_refs: list[str | None] = []
        self.foreign_keys: list[bool] = []
        self.primary_keys: list[bool] = []
        self.unique: list[bool] = []

        self.primary_key: int = -1
        self.all_srcref: str | None = source_ref

    def __str__(self) -> str:
        st = "MAPPER : Type " + self.type + " ------------\n"
        for key, val, ds, de, dp, src in zip(self.keys, self.values, self.date_start, self.date_end,
                                             self.date_point, self.source_refs):
            st += key + " --> " + str(val)

            if ds is not None or de is not None:
                st += f" [{ds} -> {de}]"
            if dp is not None:
                st += f" [{dp}]"
            if src is not None:
                st += f" {{src={src}}}"
            if self.all_srcref is not None:
                st += f" {{src={self.all_srcref}}}"
            st += "\n"
        st += "--------------------------\n"
        return st

    def __len__(self) -> int:
        return len(self.keys)

    def get_dict_key(self, key: str) -> dict[str, str] | None:
        """Récupère le dictionnaire contenant les informations d'une clé.

        Parameters:
            key: La clé, qui est également le nom de la propriété de l'objet.

        Returns:
            Si la clé éxiste, le dictionnaire contenant les informations liées à
                la clé, renvoyé par [get_dict_i][mapper.Mapper.get_dict_i].
                Sinon, renvoie `None`

        Raises:
            ValueError: Plusieurs clés correspondent.
        """
        c = self.keys.count(key)
        if c == 0:
            return None
        elif c == 1:
            return self.get_dict_i(self.keys.index(key))
        else:
            raise ValueError(f"Several corresponding key '{key}'"
                             "have been found. Only one was expected.")

    def get_dict_i(self, index: int) -> dict[str, Any]:
        """Renvoie le dictionnaire correspondant à un index.

        Parameters:
            index: L'index correspondant aux informations que l'on souhaite
                récupérer.

        Returns:
            Un dictionnaire contenant:

                - `key`: Le nom complet de la propriété.
                - `value`: La valeur de la propriété.
                - `type_value`: *À implémenter*
                - `date_start`: Date de début de la propriété.
                - `date_end`: Date de fin de la propriété.
                - `date_point`: Date de la propriété.
                - `unique`: Si la propriété, doit être unique pour l'objet.
                - `source_ref`: La source de référence de la propriété.
        """
        dic: dict[str, Any] = {}
        dic["key"] = self.keys[index]
        dic["value"] = self.values[index]
        dic["type_value"] = None  # :todo: Implementation ???
        dic["date_start"] = self.date_start[index]
        dic["date_end"] = self.date_end[index]
        dic["date_point"] = self.date_point[index]
        dic["unique"] = self.unique[index]
        if self.source_refs[index] is not None:
            dic["source_ref"] = self.source_refs[index]
        elif self.all_srcref is not None:
            dic["source_ref"] = self.all_srcref
        else:
            dic["source_ref"] = None
        return dic

    def get(self,
            key: str,
            default: Any | None = None) -> Any:
        """Retourne la valeur associée à une clé.

        Parameters:
            key: Clé à rechercher.
            default: Valeur à retourner si la clé n'est pas connue.

        Note: FutureDev
            Uniquement la première valeur du tableau. Implémenter la gestion de clefs multiples
        """
        try:
            return self.values[self.keys.index(key)]
        except ValueError:
            return default

    def get_keys_index(self,
                       key: str,
                       startswith: bool = False,
                       alias: bool = False,
                       exact: bool = True) -> list[int]:
        """Récupère la liste des indexs correspondants à la clé.

        Parameters:
            key: La clé à rechercher.
            startswith:
            alias:
            exact:

        Returns:
            La liste des indexs correspondants.
        """
        lst_key: list[int] = []
        for index, k in enumerate(self.keys):
            if k == key and exact:
                lst_key.append(index)
                continue
            if startswith and k.startswith(key + "@"):
                lst_key.append(index)
                continue
            spec_word = ["Code", "Name", "CodeAlias", "NameAlias"]
            tbl_spec = [key + w for w in spec_word]
            if startswith and (k.split("@")[0] in tbl_spec):
                lst_key.append(index)
                continue
            if alias and k == key + "Alias" and exact:
                lst_key.append(index)
                continue
            if alias and startswith and k.startswith(key + "Alias@"):
                lst_key.append(index)
                continue
        return lst_key

    def add(self, key: str,
            value: str | MTObject | Geometry,
            date_point: str | date | datetime | None = None,
            date_start: str | date | datetime | None = None,
            date_end: str | date | datetime | None = None,
            year: str | int | None = None,
            month: str | int | None = None,
            source_ref: str | None = None,
            primary_key: bool = False,
            foreign_key: bool = False,
            unique: bool = False,
            ) -> None:
        """Ajoute une propriété au Mapper.

        Parameters:
            key: Le nom de la propriété.
            value: La valeur de la propriété.
            date_point: La date de la propriété.
            date_start: La date de début de la propriété.
            date_end: La date de fin de la propriété.
            year: L'année de la propriété.
            month: Le mois de la propriété.
            source_ref: La source de référence de la propriété.
            primary_key: Indique si la propriété est une
                [clé primaire](../quickstart/indepth.md#cles-primaires-et-etrangeres)
                (uniquement pour les `Code` ou `Name`)
            foreign_key: Indique si la propriété est une
                [clé étrangère](../quickstart/indepth.md#cles-primaires-et-etrangeres)
                (uniquement pour les `Code` ou `Name`)
            unique: Vérifie l'unicité de la propriété avant de l'associer au MTObject
                (En développement).
        """
        # To depreciate
        if key == "SourceRef":
            raise ValueError("SourceRef is not an available key at the moment.")

        if key.startswith("Primary:") or primary_key:
            if key.startswith("Primary:"):
                key = key.replace("Primary:", "", 1)
                warnings.warn("Deprecated use of 'Primary:' in direct Mapper args name."
                              " Use attribute 'primary_key=' instead")
                primary_key = True
            self.primary_key = len(self)

        if (primary_key or foreign_key) and not key.startswith(('Code', 'Name')):
            raise ValueError("`primary_key` can only be associated with Code or Name")
        if primary_key and foreign_key:
            raise ValueError('Cannot have `primary_key` and `foreign_key`'
                             ' at the same time.')

        date_point = to_date(date_point)
        start_n, end_n = get_start_end_dates(start=date_start, end=date_end,
                                             year=year, month=month)
        if (start_n, end_n) == get_start_end_dates():
            start_n, end_n = None, None

        if start_n is not None and end_n is not None and date_point is not None:
            raise AttributeError("Too many information about date (start, end)+(point)")

        self.keys.append(key)
        self.values.append(value)
        self.date_start.append(start_n)
        self.date_end.append(end_n)
        self.date_point.append(date_point)
        self.source_refs.append(source_ref)
        self.unique.append(unique)
        self.primary_keys.append(primary_key)
        self.foreign_keys.append(foreign_key)

    def is_foreign_key(self, keyindex: int) -> bool:
        """Détermine si la propriété à un certain index est une clé étrangére."""
        return self.foreign_keys[keyindex]

    def is_primary_key(self, keyindex: int) -> bool:
        """Détermine si la propriété à un certain index est une clé primaire."""
        return self.primary_key == keyindex


class MapperError(Exception):
    def __init__(self,
                 mapper: Mapper,
                 result: dict[str, str],
                 key: str
                 ) -> None:
        self.mapper = mapper
        self.result = result
        self.key = key
        print("--------- ERROR CONTEXT : MAPPER ----------------")
        print(mapper)
        print("--------- ERROR CONTEXT : RESULT ----------------")
        print(result)
        print()
        try:
            msg = "Wrong data associated with key '%s', error info = %s" % (key, result[key])
        except KeyError:
            msg = "Unknown error '%s'" % (key)
        super().__init__(msg)



import time

info_timer = time.time()


def info(string):
    global info_timer
    delta_time = time.time() - info_timer
    print(f"/\\/\\/\\/ {string} ({delta_time:.3f} s)")
    info_timer += delta_time



from .mtobject import MTObject
from .quantifiable import MTQuantifiable

from .territory import Territory
from .actor import Actor
from .product import Product
from .pathflow import Pathflow
from .gateflow import Gateflow
from .stock import Stock
from .property import Property



import math
import sys

standard_unit_energy = ["Wh", "J", "tep", "WhEP", "WhEF", "J_PCS", "J_PCI", "Wh_EP", "Wh_EF", "Wh_PCI", "Wh_PCS", "W"]
standard_unit_mass = ["g", "t"]
standard_unit_surface = ["m2", "a"]
standard_unit_volume = ["m3", "L"]
standard_unit_money = ["euro", "€"]
standard_unit_other = ["geqCO2", "teqCO2", "g(P)", "g(N)", "g(O2)", "g/l"]
standard_unit = standard_unit_energy + standard_unit_mass +\
                standard_unit_surface + standard_unit_volume +\
                standard_unit_money + standard_unit_other

## PREFIX DU SYSTEM INTERNATIONAL
prefix_si = {'y':1e-24, 'z':1e-21, 'a':1e-18} # yocto zepto atto

prefix_si['f'] = 10**(-15) # femto
prefix_si['p'] = 10**(-12) # pico
prefix_si['n'] = 10**(-9) # nano
prefix_si['u'] = 10**(-6) # micro
prefix_si['µ'] = 10**(-6) # micro
prefix_si['m'] = 10**(-3) # milli
prefix_si['c'] = 10**(-2) #centi
prefix_si['d'] = 10**(-1) #deci

prefix_si['D'] = 10**1 #deca  => ! D != da
prefix_si['h'] = 10**2 #hecto
prefix_si['k'] = 10**3
prefix_si['M'] = 10**6
prefix_si['G'] = 10**9
prefix_si['T'] = 10**12
prefix_si['P'] = 10**15
prefix_si['E'] = 10**18
prefix_si['Z'] = 10**21
prefix_si['Y'] = 10**24

prefix_si_keys = list(prefix_si.keys())

std_cst_coeff_dic = {} # Dictionnaire des coefficients constants standards

std_cst_coeff_dic["J/Wh"] = 3600.0
std_cst_coeff_dic["J/Wh_PCI"] = 3600.0
std_cst_coeff_dic["Wh/Wh_PCI"] = 1
std_cst_coeff_dic["J_PCI/Wh_PCI"] = 3600.0
std_cst_coeff_dic["J_PCS/Wh_PCS"] = 3600.0
std_cst_coeff_dic["J/tep"] = 41.868 * (10 ** 9)
std_cst_coeff_dic["L/m3"] = 1000.0
std_cst_coeff_dic["m2/a"] = 100.0
std_cst_coeff_dic["euro/€"] = 1
std_cst_coeff_dic["g/t"] = 1000000.0
std_cst_coeff_dic["geqCO2/teqCO2"] = 1000000.0
std_cst_coeff_dic["gCO2e/tCO2e"] = 1000000.0
std_cst_coeff_dic["m³/m3"] = 1
std_cst_coeff_dic["J/cal"] = 4.184

class ConversionError(Exception):
    pass

class UnitConverter():

    def __init__(self, dico):
        self.unitset = set()
        self.unitdict = {}

        for key, val in dico.items():
            self.add_conversion(key, val)

    def add_conversion(self, conversionkey, value):
        mysplitkey = conversionkey.split("/")
        c1u, c1c = get_unit_coeff(mysplitkey[0])
        c2u, c2c = get_unit_coeff(mysplitkey[1])

        if c1u not in self.unitdict:
            self.unitdict[c1u] = {}
        if c2u not in self.unitdict[c1u]:
            self.unitdict[c1u][c2u] = value*c1c/c2c

        if c2u not in self.unitdict:
            self.unitdict[c2u] = {}
        if c1u not in self.unitdict[c2u]:
            if value != 0:
                self.unitdict[c2u][c1u] = c2c/(c1c*value)
            else:
                self.unitdict[c2u][c1u] = sys.float_info.max

    def do(self, from_unit, to_unit, quantity, extendlimit=5):
        fu, fc = get_unit_coeff(from_unit)
        tu, tc = get_unit_coeff(to_unit)
        #print("Conversion = ", fu, fc, tu, tc)
        if fu == tu:
            return quantity * float(fc) / float(tc)

        while extendlimit > 0:
            coeff = self.get_fromto_coeff(fu, tu)
            if coeff is not None:
                if coeff == 0:
                    return sys.float_info.max
                else:
                    return quantity * float(fc) / (float(tc) * coeff)
            self.extend_converter()
            #print(self)
            #print("(E) --------")
            extendlimit -= 1

        error_message = "Undefined conversion %s %s --> %s\nCoeff dict = %s" % \
                        (quantity, from_unit, to_unit, "")

        """print("################## ERROR INFO #################################################################")
        print(error_message)
        print(self.unitdict)"""
        raise ConversionError(error_message)


    def get_fromto_coeff(self, fu, tu):
        if fu in self.unitdict:
            if tu in self.unitdict[fu]:
                return self.unitdict[fu][tu]
        return None

    def extend_converter(self):
        extraliner = []
        for c1a, dico in self.unitdict.items():
            for c2a, vala in dico.items():
                if c2a in self.unitdict:
                    for c2b, valb in self.unitdict[c2a].items():
                        extraliner.append(("%s/%s"%(c1a, c2b), vala*valb))

        for line in extraliner:
            self.add_conversion(*line)

    def __str__(self):
        mystr = ""
        for key1, dico in self.unitdict.items():
            for key2, val in self.unitdict[key1].items():
                mystr += "%s/%s  =  %s\n" % (key1, key2, val)
        return mystr


def extend_norm_coeffdic(coeffdic):
    """ Autocombine un dictionnaire pour extraire les coefficients"""
    extendcoeffdic = coeffdic.copy()
    for key1, val1 in coeffdic.items():
        mycoeffsplit = key1.split("/")
        if len(mycoeffsplit) != 2:
            raise AttributeError("Uncorrect coefficient '%s'" % val1)
        c1, c2 = mycoeffsplit[0], mycoeffsplit[1]
        for key2, val2 in coeffdic.items():
            mycoeffsplit = key2.split("/")
            d1, d2 = mycoeffsplit[0], mycoeffsplit[1]
            if c2 == d1:
                mykey = "%s/%s" % (c1, d2)
                if mykey not in extendcoeffdic:
                    extendcoeffdic[mykey] = val1*val2
            if c1 == d2:
                mykey = "%s/%s" % (d1, c2)
                if mykey not in extendcoeffdic:
                    extendcoeffdic[mykey] = val1*val2

    return {**coeffdic, **extendcoeffdic}


def get_unit_coeff(unit_name):
    """ Renvoie un tuple (Unité, coeff)
    Ex : MWh -> (Wh, 10^6)"""

    if unit_name is None:
        return (None, None)

    if unit_name[0] in prefix_si_keys and unit_name[1:] in standard_unit:
        # GESTION DES UNITE ^2 et ^3 (m2 et m3 pour l'instant)
        tempcoeff = prefix_si[unit_name[0]]
        if unit_name[1:] == "m2":
            tempcoeff = prefix_si[unit_name[0]]**2
        if unit_name[1:] == "m3":
            tempcoeff = prefix_si[unit_name[0]]**3
        return (unit_name[1:], tempcoeff)
    else:
        return (unit_name, 1)


def convert(quantity, from_unit, to_unit, coeffdic=None):
    """
    Converti une valeur d'une unité vers une autre.

    Parameters
        quantity: Quantité initiale
        from_unit: Unité initiale
        to_unit: Unité finale
        coeffdic: Dictionnaire complémentaire (Optionnel)

    Examples:
        >>> from sinamet.tools.unitconverter import convert
        >>> convert(12, 't', 'kg') # Converti des tonnes en kilo.
        >>> convert(12000, 'L', 'm3') # Converti des tonnes en kilo.
        >>> convert(12, 't', 'kg') # Converti des tonnes en kilo.
        >>> convert(1378, 'kg', 'm3', coeffdic={"g/cm3":7.874}) # Utilise un coefficient de densité massique complémentaire
    """
    if coeffdic is None:
        coeffdic = std_cst_coeff_dic
    else:
        coeffdic = {**std_cst_coeff_dic, **coeffdic}
    uc = UnitConverter(coeffdic)
    #print(uc)
    return uc.do(from_unit, to_unit, quantity)


####### TESTS ##################"
def test_maker(qte, fu, tu, expected, coeffdic = None):
    if coeffdic is None:
        coeffdic = std_cst_coeff_dic
    computed = convert(qte, fu, tu, coeffdic=coeffdic)
    print("%s %s = %s %s [check=%s]" % (qte, fu, computed, tu, expected), end="")
    if abs(computed - expected) > computed/100:
        print(" -> Failed")
    else:
        print(" -> Passed")


if __name__ == "__main__":
    print("STARTING TEST UNIT CONVERTER")
    #test_maker(10, "MWh", "kWh", 10000)
    test_maker(1234, "MW", "kW", 1234000)
    test_maker(36000, "J", "Wh", 10)
    test_maker(36, "kJ", "kWh", 0.01)
    test_maker(1, "m3", "L", 1000)
    test_maker(1, "km3", "L", 1000000000000)
    test_maker(2, "ha", "m2", 20000)
    test_maker(11630, "kWh", "tep", 1)

    # Gazole test
    mygazole_dict = {"kWh/kg": 11.675167030567687, "kgeqCO2/kg":3.85}
    test_maker(455991410, "kg", "MWh", 5323775, coeffdic=mygazole_dict)

    print("ALL TESTS PASSED")



class SidbError(Exception):
    """Exception de base pour les erreurs Sinamet."""
    pass


class SidbNotFoundError(SidbError):
    """L'objet requété n'a pas pu être trouvé."""
    pass


class SidbMultiFoundError(SidbError):
    """Plus d'un objet correspondant à la requète à été trouvé."""
    pass



from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship

from .quantifiable import MTQuantifiable
from .actor import Actor
from .territory import Territory


class Stock(MTQuantifiable):
    """Table stock."""
    __tablename__ = 'stock'

    id: Mapped[int] = mapped_column(ForeignKey('mtquantifiable.id', ondelete='CASCADE'),
                                    primary_key=True)

    actor_id: Mapped[int | None] = mapped_column(ForeignKey('actor.id'))
    actor: Mapped[Actor | None] = relationship()

    territory_id: Mapped[int | None] = mapped_column(ForeignKey('territory.id'))
    territory: Mapped[Territory | None] = relationship()

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }



from sqlalchemy.schema import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column

from .mtobject import MTObject


class Territory(MTObject):
    """Table territoire."""
    __tablename__ = 'territory'

    id: Mapped[int] = mapped_column(ForeignKey('mtobject.id', ondelete='CASCADE'),
                                    primary_key=True)

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }

    def __str__(self) -> str:
        return f'<Territory-{self.id}: {self.get_code()}-{self.get_name()}>'



from datetime import datetime, date, timedelta
from typing import overload

import pandas


def full_date_range(start, end, freq):
    """Return TimeSeries, with frequency"""
    _start_date = start
    _end_date = end  # END DATE INCLUDE IN TIME STEP !

    # Building list of dates
    # Nearest date (with freq) before start
    _startlistdates = pandas.date_range(end=_start_date,
                                        freq=freq,
                                        periods=1,
                                        closed='right')
    # Nearest date (with freq) after start
    _endlistdates = pandas.date_range(start=_end_date,
                                      freq=freq,
                                      periods=1,
                                      closed='left')
    # All dates in between
    _listdates = pandas.date_range(start=_start_date,
                                   end=_end_date,
                                   freq=freq)

    # print(_startlistdates)
    # print(_endlistdates)
    # print(_listdates)

    rslt = pandas.DatetimeIndex.union(_startlistdates, _listdates)
    rslt = pandas.DatetimeIndex.union(rslt, _endlistdates)
    return rslt

def common_nb_days(start1, end1, start2, end2):
    """End date excluded"""
    if start1 <= start2 and end1 >= end2:
        _return = (end2 - start2).days
    elif start2 <= start1 and end2 >= end1:
        _return = (end1 - start1).days
    elif end1 <= start2: # Period 1 before period 2
        _return = 0
    elif start1 >= end2:
        _return = 0
    elif start1 <= start2 and end1 <= end2:
        _return = (end1 - start2).days
    elif start1 >= start2 and end1 >= end2:
        _return = (end2 - start1).days
    else:
        raise ValueError("Unable to perform common_nb_days")
    #print(str(start1) + " -> " + str(end1) + " // " + \
    #      str(start2) + " -> " + str(end2) + " = " + str(_return))
    return _return


@overload
def to_date(d:  None) -> None: ...
@overload
def to_date(d: str | datetime | date) -> date: ...


def to_date(d: str | datetime | date | None) -> date | None:
    if d is None:
        return None

    if isinstance(d, str):
        try:
            return datetime.strptime(d, "%d/%m/%Y").date()
        except ValueError:
            try:
                return datetime.strptime(d, "%Y-%m-%d").date()
            except ValueError:
                raise ValueError(f"Invalid date format: '{date}'")
    elif isinstance(d, datetime):
        return d.date()
    elif isinstance(d, date):
        return d
    else:
        raise TypeError(f"Unknown date type: '{type(date)=}'")


def get_start_end_dates(start: str | datetime | date | None = None,
                        end: str | datetime | date | None = None,
                        year: int | str | None = None,
                        month: int | str | None = None
                        ) -> tuple[date, date]:
    """
        - start => start -- date.max
        - end => date.min -- end
        - start & end => start -- end
        - year => 01/01/year -- 12/31/year
        - year & month => 01/month/year -- 01/month+1/year - 1
        - no infos => date.min -- date.max
    """
    if month and year is None:
        raise ValueError("Cannot have a month without a year.")

    start, end = map(to_date, (start, end))
    if year is not None:
        if start or end:
            raise ValueError("Invalid arguments: cannot have a year with a start"
                             " or an end.")
        if isinstance(year, str):
            year = int(year)
        start = date(year, 1, 1)
        end = date(year, 12, 31)
        if month is not None:
            if isinstance(month, str):
                month = int(month)
            start = start.replace(month=month)
            end = (date(year + (month == 12), month % 12 + 1, 1)
                   - timedelta(days=1))
    if not start:
        start = date.min
    if not end:
        end = date.max
    return start, end


def timedelta_coefficient(start1, end1, start2, end2):
    timedelta1 = (end1 - start1 + timedelta(days=1)).days
    timedelta2 = (end2 - start2 + timedelta(days=1)).days
    return (float(timedelta2) / float(timedelta1))


def get_month_list(lang='fr'):
    if lang == 'fr':
        return ["Janvier", "Février", "Mars", "Avril", "Mai", "Juin",
                "Juillet", "Août", "Septembre", "Octobre", "Novembre", "Décembre"]
    else:
        raise ValueError("Unknown langage : '%s'"%lang)


def get_month_name(num, lang='fr'):
    return get_month_list(lang=lang)[num-1]


def to_jstimestamp(mydatetime: str | datetime | date | None) -> float | None:
    mydatetime = to_date(mydatetime)
    if not mydatetime:
        return None

    timestamp_start = (datetime.combine(mydatetime, datetime.min.time())
                       - datetime(1970, 1, 1)) / timedelta(seconds=1)
    return int(timestamp_start) * 1e3  # Multiply by 1e3 as JS timestamp is in milliseconds

# MS = mois ; AS = an ; W = semaine ; D = jour



from __future__ import annotations

from typing import Literal

from datetime import date

from sqlalchemy import ForeignKey
from sqlalchemy import PickleType
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship

from sinamet.tools.unitconverter import convert, get_unit_coeff, ConversionError

from .mtobject import MTObject
from .product import Product


class MTQuantifiable(MTObject):
    """
    Classe mère des objets quantifiables du modèle de Sinamet : `Gateflow`, `Pathflow`, `Stock`

    Parameters:
        properties: Liste des propriétés de l'objet
        type: type d'objet (parmis `gateflow`, `pathflow`, `stock`)
        product: Produit associé à l'élément quantifiable
        quantity: Dictionnaire de la quantité dans différentes unités
        date_start : Repère temporel (Début) de l'élément quantifiable
        date_end: Repère temporel (Fin) de l'élément quantifiable
        date_point: Repère temporel (Date ponctuelle) de l'élément quantifiable
    """
    __tablename__ = 'mtquantifiable'
    __mapper_args__ = {
        'polymorphic_identity': __tablename__
    }

    id: Mapped[int] = mapped_column(ForeignKey('mtobject.id', ondelete='CASCADE'),
                                    primary_key=True)

    product_id: Mapped[int | None] = mapped_column(ForeignKey('product.id'))
    product: Mapped[Product | None] = relationship(foreign_keys=product_id)

    date_point: Mapped[date | None]

    quantity: Mapped[PickleType] = mapped_column(PickleType)

    def __init__(self):
        MTObject.__init__(self)
        self.quantity = {}

    def __str__(self) -> str:
        quantity = next((f'{val} {key}' for key, val in self.quantity.items()), None)

        if self.date_point is not None:
            date = f'[{self.date_point}]'
        else:
            date = f'[{self.date_start} -> {self.date_end}]'

        if (name := self.get_name()) is not None:
            name = f'Name: {name}'
        if (code := self.get_code()) is not None:
            code = f'Code: {code}'
        product_name = str(self.product.get_name()) if self.product else ''

        return f'<{" | ".join(filter(None, (
            f'{self.__class__.__name__}-{self.id}',
            name, code, product_name, date, quantity
        )))}>'

    def get_quantity(self, unit: str,
                     error: Literal['raise', 'none', 'print'] = 'raise',
                     ) -> float | None:
        """Retourne une quantité.

        Parameters:
            unit: Unité de la quantité (euro, kg, ...).
            error: Comportement à adopter en cas d'erreur.

        Note: FutureDev
            Remasteriser cette fonction.
        """
        try:
            product = self.product
            if product is None:
                raise AttributeError("None product associated to : %s" % repr(self))

            quant = self.quantity
            # CAS 1 : Unité déjà présente dans quant : on retourne
            if unit in quant:
                return quant[unit]

            # CAS 2 : Unité déjà présente dans quant avec un prefixe différent ou
            # une équivalence prédéterminée
            for key, val in quant.items():
                try:
                    return convert(val, key, unit)
                except ConversionError:
                    pass

            # CAS 3 : Unité non directement convertible, on utilise les coefficients
            # Associés au produit
            # Construction du dictionnaire
            lst_coeff_attr = product.get_properties("Coefficient",
                                                    return_type="property")
            if not lst_coeff_attr:
                raise ConversionError(f'Unable to convert: no coefficients found in {product=}'
                                      f' ({quant=})')

            # Extracting information
            if type(lst_coeff_attr) is not list:
                lst_coeff_attr = [lst_coeff_attr]

            _conversiondict = {}
            for coeff_attr in lst_coeff_attr:
                _coeff = coeff_attr.name_b
                # _conversiondict[_coeff] = coeff_attr.get_value() TODO: à vérifier

                (_unit1, _unit2) = _coeff.split("/")
                # Correction des coefficients (parce qu'on peut avoir kWh / Euros)
                (standard_unit1, coeff_unit1) = get_unit_coeff(_unit1)
                (standard_unit2, coeff_unit2) = get_unit_coeff(_unit2)
                _conversiondict[standard_unit1 + "/" + standard_unit2] = (coeff_attr.get_value()
                                                                          * float(coeff_unit1)/coeff_unit2)
            for key, val in quant.items():
                try:
                    return convert(val, key, unit, _conversiondict)
                except ConversionError:
                    pass
            # CAS 4 : ???
            raise ConversionError("Unable to express : %s in %s. Available coefficients = %s" % (self.quantity, unit, _conversiondict))

        except ConversionError as e:
            if error == "none":
                return None
            print("ERROR WITH = %s" % repr(self))
            if error == "print":
                print(e)
                return None
            raise e

    def copy(self) -> MTQuantifiable:
        """Deep copy d'un objet quantifiable.

        Note: FutureDev
            Vérifier usage.
        """
        mtq = self.__class__()
        mtq.product = self.product
        mtq.quantity = self.quantity.copy()
        return mtq



import pandas as pd
from datetime import timedelta

def list_to_dataframe(list_mtobject, unit=None, per_day = True, uniterror="raise"):
    if len(list_mtobject) == 0:
        return pd.DataFrame()
    typobj = list_mtobject[0].__class__.__name__.lower()

    if typobj not in ["pathflow", "gateflow", "stock", "territory", "actor"]:
        raise TypeError("Unknown list object type : %s" % typobj)

    df = pd.DataFrame.from_dict({'mtobject': list_mtobject})

    if (unit is None) and (uniterror=="none"):
        return df

    if typobj in ["pathflow", "gateflow"]:
        df = dfbuild_timeperiod(df)

    if typobj in ["pathflow", "gateflow", "stock"]:
        if type(unit) is not list:
            unit = [unit]
        for myunit in unit:
            if myunit == 'x':
                df[myunit] = 1
            else:
                df[myunit] = df['mtobject'].apply(lambda v: v.get_quantity(myunit, error=uniterror))
                if per_day:
                    try:
                        df[myunit + "_per_day"] = df[myunit] / df['nb_days']
                    except TypeError as e:
                        if uniterror == "none":
                            pass
                        else:
                            raise TypeError(str(e) + " : You should maybe set 'per_day' to False")
                    except KeyError as e:
                        if uniterror == "none":
                            pass
                        else:
                            raise e
    return df

def dfbuild_timeperiod(df):
    # LIST FLOWS = DATE START + END INCLUES
    # DATAFRAME = START INCLUE, END EXCLUE
    df['date_start']= df['mtobject'].apply(lambda v: date_point_flow_manage(v, "start"))
    df['date_end'] = df['mtobject'].apply(lambda v: date_point_flow_manage(v, "end"))
    """df['date_start'] = df['mtobject'].apply(lambda v: v.date_start)
    df['date_end'] = df['mtobject'].apply(lambda v: v.date_end + timedelta(days=1))"""
    df['nb_days_timedelta'] = df['date_end'] - df['date_start']
    df['nb_days'] = df['nb_days_timedelta'].apply(lambda v: v.days)

    return df

def date_point_flow_manage(mtqobject, dir):
    if mtqobject.date_point is not None:
        dst = mtqobject.date_point
        den = mtqobject.date_point + timedelta(days=1)
    else:
        dst = mtqobject.date_start
        den = mtqobject.date_end + timedelta(days=1)
    if dir == "start":
        return dst
    elif dir == "end":
        return den